#############################################################################
##
#W  detdesc.gi                                                   Bettina Eick
##
if not IsBound( CHECK ) then CHECK := false; fi;
if not IsBound( FULLT ) then FULLT := true; fi;

#############################################################################
##
#F BasicCommutators()
##
Cinv := function( a, b ) return a * b * a^-1 * b^-1; end;
BasicCommutators := function( a, b )
    return [a, b, Cinv( a, b ), Cinv( Cinv( b, a ), b ), 
            Cinv( b, Cinv( a, Cinv( b, b ) ) ), 
            Cinv( b, Cinv( a, Cinv( b, Cinv( b, b ) ) ) ), 
            Cinv( Cinv( b, Cinv( a, b ) ), Cinv( b, a ) ) ];
end;

#############################################################################
##
#F AllPositions( list, e )
##
AllPositions := function( list, e )
    return Filtered( [1..Length(list)], x -> list[x] = e );
end;
    
#############################################################################
##
#F DualRowSpace( V, d, F )
##
DualRowSpace := function( V, d, F )
    local l, n, r;
    if Length( V ) = 0 then return IdentityMat( d, F ); fi;
    l := Length( V );
    n := Length( V[1] );
    for r  in [ l + 1 .. n ]  do Add( V, 0*V[1] ); od;
    return NullspaceMat( TransposedMat( V ) );
end;

#############################################################################
##
#F MaximalSubgroupsPGroup( G )
##
MaximalSubgroupsPGroup := function( G )
    local p, P, g, d, f, I, irr, max, i;
    p := PrimePGroup(G);
    P := PRump( G, p );
    g := Pcgs(G) mod Pcgs(P);
    d := Length( g );
    f := GF( p );
    irr := NormedVectors( f^d );
    max := List( irr, x -> DualRowSpace([x], d, f) );
    for i in [1..Length(max)] do
        max[i] := List( max[i], x -> MappedVector( IntVecFFE(x), g ) );
        max[i] := InducedPcgsByPcSequenceAndGenerators(Pcgs(G),Pcgs(P),max[i]);
        max[i] := SubgroupByPcgs( G, max[i] );
    od;
    return max;
end;

#############################################################################
##
## check invariants of low index subgroups
##
IsCovered := function( sub, inv )
    local j, i;
    j := 0;
    for i in [1..Length(sub)] do
        j := j + 1;
        if j > Length(inv) then return false; fi;
        while not IsInt( inv[j]/sub[i] ) do
            j := j + 1; 
            if j > Length(inv) then return false; fi;
        od;
    od;
    return true;
end;

CheckMaximalSubgroups := function( G, inv )
    local max;
    max := MaximalSubgroupsPGroup( G );
    max := List( max, AbelianInvariants );
    max := Filtered( max, x -> IsCovered( x, inv ) );
    return Length(max) >= 1;
end;

#############################################################################
##
#F InvariantsSubgroup( G, l )
##
InvariantsSubgroup := function( G, l )
    local max, inv, i;
    max := [G];
    inv := [];
    for i in [1..l] do
        max := Set( Flat( List( max, x -> MaximalSubgroupsPGroup(x) ) ) );
        inv[i] := List( max, AbelianInvariants );
        Sort( inv[i] );
    od;
    return inv;
end;
        
#############################################################################
##
## ConjugatingElement
##
ConjugatingElement := function( gensK, expsC, gensC, a, b, w )
    local k, s;
    k := ExponentsOfPcElement( gensK, a * b^-1 ) * w;
    s := SolutionMat( expsC, k );
    if not IsBool(s) then 
        s := MappedVector( IntVecFFE(s), gensC );
    fi;
    return s;
end;

#############################################################################
##
## LiftClasses( G, H, prs )
##
LiftClasses := function( G, H, prs )
    local p, w, gensH, gensG, n, l, imgsH, gensK, baseK, vecsK, new,
          i, j, k, c, s, gensC, imgsC, expsC, triaC, R, reps, r, h, t;

    # set up
    p := PrimePGroup( G );
    w := One( GF( p ) );

    # get H and G
    gensH := Pcgs(H);
    gensG := Pcgs(G);
    n := Length( gensH );
    l := Length( gensG );

    # get mapping H -> G and kernel
    imgsH := gensH{[1..l]};
    gensK := InducedPcgsByPcSequenceNC( gensH, gensH{[l+1..n]} );
    baseK := IdentityMat( n-l, GF(p) );
    vecsK := GF(p)^(n-l);

    # lift G!.cls to H.cls
    H!.cls := [];
    for i in [1..Length(G!.cls)] do
        new := [];
        if i = 2 then p := 5; else p := prs; fi;
        for j in [1..Length(G!.cls[i])] do

            # get class
            c := G!.cls[i][j];
            s := MappedPcElement( c.repr, gensG, imgsH );

            # compute exponents of commutators
            gensC := List( c.cent, x -> MappedPcElement(x, gensG, imgsH) );
            imgsC := List( gensC, y -> Comm(s,y) );
            expsC := List( imgsC, y -> ExponentsOfPcElement( gensK, y )*w );
            triaC := EcheloniseMat(expsC);

            # compute centralizer and representatives
            if Length(triaC) = 0 then
                R := ShallowCopy(gensC);
                reps := Elements( vecsK );
            else
                R := ShallowCopy( TriangulizedNullspaceMat( expsC ) );
                R := List( R, IntVecFFE );
                R := List( R, x -> MappedVector( x, gensC ) );
                reps := BaseSteinitzVectors( baseK, triaC ).factorspace;
                reps := Elements( Subspace(vecsK, reps) );
            fi;
            Append( R, AsList(gensK) );
            reps := List( reps, y -> s * MappedVector( IntVecFFE(y), gensK ) );

            # the complex conjugation
            if i = 1 then
                for r in reps do
                    if Order(r) = 2 then
                        Add( new, rec( repr := r, cent := R ) );
                    fi;
                od;
            fi;

            # the others
            if i > 1 then
                h := MappedPcElement( c.conj, gensG, imgsH );
                for r in reps do
                    t := ConjugatingElement(gensK, expsC, gensC, r^p, r^h, w);
                    if not IsBool( t ) then
                        Add( new, rec( repr:=r, cent:=R, conj:=t*h) );
                    fi;
                od;
            fi;

            # reduce by odd powers
            if i = 3 then 
                
            fi;
        od;
        if Length( new ) = 0 then return false; fi;
        H!.cls[i] := new;

        # do a check
        if CHECK then 
            Print("checking ... \n");
            new := List( new, x -> x.repr );
            for j in [1..Length(new)] do
                for k in [j+1..Length(new)] do
                    s := RepresentativeAction( H, new[j], new[k] );
                    if not IsBool(s) then Error("hier"); fi;
                od;
            od;
            Print("... done \n");
        fi;
        
    od;
    return true;
end;

#############################################################################
##
#F CheckClasses( G, prs )
##
CheckClasses := function( G, prs )
    local c;
    
    for c in G!.cls[1] do 
        if Order( c.repr ) <> 2 then 
            Print("complex conjugation wrong \n");
            return false; 
        fi;
    od;
    for c in G!.cls[2] do
        if not c.repr^5 = c.repr^c.conj then
            Print("5 wrong \n");
            return false; 
        fi;
    od;
    for c in G!.cls[2] do
        if not c.repr^prs = c.repr^c.conj then 
            Print("prime wrong \n");
            return false; 
        fi;
    od;
    return true;
end;

#############################################################################
##
#F CheckRanks( D )
##
CheckRanks := function( D )
    return MultiplicatorRank(D) - NuclearRank(D) <= 2;
end;
    
#############################################################################
##
#F RestrictedDescendantsRec( G, cl, prs, inv )
##
RestrictedDescendantsRec := function( G, cl, prs, inv )
    local d, i;

    d := PqDescendants( G : AllDescendants := FULLT );
    if Length(d) = 0 then return G; fi;

    Info( InfoDetDesc, 2, "  class ",cl,
                          " of size 2^", Length(Factors(Size(G))),
                          " got ", Length(d)," descendants");
    for i in [1..Length(d)] do
        Info( InfoDetDesc, 3, "    checking number ",i);

        # check  multiplicator and nuclear rank
        if not IsBool( d[i] ) then 
            Info( InfoDetDesc, 3, "    by ranks");
            if not CheckRanks( d[i] ) then 
                d[i] := false;
            fi;
        fi;

        # check cls and add D.cls
        if not IsBool( d[i] ) then 
            Info( InfoDetDesc, 3, "    by classes");
            if not LiftClasses( G, d[i], prs ) then 
                d[i] := false;
            fi;
        fi;

        # check invs
        if not IsBool( d[i] ) then 
            Info( InfoDetDesc, 3, "    by maximal subgroups");
            if not CheckMaximalSubgroups( d[i], inv ) then 
                d[i] := false;
            fi;
        fi;

        if not IsBool( d[i] ) then 
            Info( InfoDetDesc, 3, "    survived");
            d[i] := RestrictedDescendantsRec( d[i], cl + 1, prs, inv );
        fi;
    od;
    return Filtered( Flat( d ), x -> not IsBool(x) );
end;

#############################################################################
##
#F RestrictedDescendantsItr( G, cl, prs, inv )
##
RestrictedDescendantsItr := function( G, cl, prs, inv )
    local todo, done, news, i, H, d;

    todo := [G];
    done := []; 
    news := [];

    while Length(todo) > 0 do
        for i in [1..Length(todo)] do
            H := todo[i];
            d := PqDescendants( H : AllDescendants := FULLT );
            if Length(d) = 0 then Add( done, H ); fi;

            Info( InfoDetDesc, 2, "  class ",cl,
                          " of size 2^", Length(Factors(Size(H))),
                          " got ", Length(d)," descendants");
            for i in [1..Length(d)] do
                Info( InfoDetDesc, 3, "    checking number ",i);

                # check  multiplicator and nuclear rank
                if not IsBool( d[i] ) then 
                    Info( InfoDetDesc, 3, "    by ranks");
                    if not CheckRanks( d[i] ) then 
                        d[i] := false;
                    fi;
                fi;

                # check cls and add D.cls
                if not IsBool( d[i] ) then 
                    Info( InfoDetDesc, 3, "    by classes");
                    if not LiftClasses( H, d[i], prs ) then 
                        d[i] := false;
                    fi;
                fi;

                # check invs
                if not IsBool( d[i] ) then 
                    Info( InfoDetDesc, 3, "    by maximal subgroups");
                    if not CheckMaximalSubgroups( d[i], inv ) then 
                        d[i] := false;
                    fi;
                fi;
            od;
            Append( news, Filtered(d, x -> not IsBool(x)) );
        od;
        cl := cl+1;
        todo := ShallowCopy(news);
        news := [];
    od;
    return done;
end;

#############################################################################
##
#F SetUpClasses( G, prs )
##
SetUpClasses := function( G, prs )
    local cls, gen, all, cl, r, e, n;

    cls := [[],[],[]];
    gen := Pcgs(G);
    all := ConjugacyClasses(G);

    # the complex conjugation
    cls[1] := [rec( repr := gen[1],
                    cent := Pcgs(Centralizer( G, gen[1] )) )];

    for cl in all do
        r := Representative(cl);
        e := ExponentsOfPcElement(gen,r);

        # check for prime 5
        if e[1] = 0 * e[1] and e[2] <> 0 * e[2] then 
            if r^5 in cl then 
                n := rec( repr := r,
                          cent := Pcgs(Centralizer(G,r)),
                          conj := RepresentativeAction(G,r,r^5) );
                Add( cls[2], n );
            fi;
        fi;

        # check for the other prime
        if e[1] <> 0 * e[1] and e[2] = 0 * e[2] then 
            if Order(r) > 2 and r^prs in cl then 
                n := rec( repr := r,
                          cent := Pcgs(Centralizer(G,r)),
                          conj := RepresentativeAction(G,r,r^prs) );
                Add( cls[3], n );
            fi;
        fi;
    od;
    
#    # the prime 5
#    cls[2] := [rec( repr := gen[2],
#                    cent := Pcgs(Centralizer( G, gen[2] )),
#                    conj := One(G) ),
#               #rec( repr := gen[2]*gen[4],
#               #     cent := Pcgs(Centralizer( G, gen[2]*gen[4] )),
#               #     conj := One(G) ),
#               rec( repr := gen[2]*gen[5],
#                    cent := Pcgs(Centralizer( G, gen[2]*gen[5] )),
#                    conj := One(G) ),
#               #rec( repr := gen[2]*gen[3]*gen[4],
#               #     cent := Pcgs(Centralizer(G, gen[2]*gen[3]*gen[4])),
#               #     conj := One(G) )
#               ];
#
#    # the other prime
#    cls[3] := [rec( repr := gen[1]*gen[4],
#                    cent := Pcgs(Centralizer( G, gen[1]*gen[4] )),
#                    conj := gen[4] )];

    G!.cls := cls;
end;

#############################################################################
##
#F GaloisCandidates( k, n, [recu] )
##
InstallGlobalFunction( GaloisCandidates, function( arg )
    local k, n, recu, F, f, a, b, r, s, G, inv, prs;

    # catch arguments
    k := arg[1];
    n := arg[2];
    if Length(arg) = 3 then recu := arg[3]; else recu := false; fi;

    # set up G / G_4
    F := FreeGroup( 2 );
    f := GeneratorsOfGroup( F ); a := f[1]; b := f[2]; 
    if n = 4 then 
        r := [a^2, b^(Comm(b,a)^2) / b^5 ];
        s := 4;
    elif n = 5 and k = 2 then 
        r := [a^2, b^(Comm(Comm(Comm(b,a),b),b)) / b^5 ];
        s := 5;
    elif n = 5 and k > 2 then 
        r := [a^2, b^(Comm(Comm(b,a),b)^2) / b^5 ];
        s := 5;
    elif n > 5 and k = 2 then 
        r := [a^2, b^(Comm(Comm(b,a),b)^2 * Comm(Comm(Comm(b,a),b),b)) / b^5 ];
        s := 5;
    elif n > 5 and k > 2 then 
        r := [a^2, b / b^5 ];
        s := 5;
    else
        Error("n must be at least 4 and k must be at least 2");
    fi;
    G := Pq( F/r : Prime := 2, ClassBound := s ); IsPGroup(G);

    # maximal subgroup invariants 
    inv := [2, 2^n];

    # primes
    if k = 0 then prs := -1; else prs := -(1+2^k); fi;

    # set up classes for G
    SetUpClasses(G, prs);

    # run determination
    if recu then 
        return RestrictedDescendantsRec( G, s, prs, inv );
    else
        return RestrictedDescendantsItr( G, s, prs, inv );
    fi;
end );

